
 /***************************************************************
 *                                                              *  
 *    o_schedule.h                                              *
 *    Time utilities                                            *
 *                                                              *
 *    (G) corTeX / Optimum   1999-02-04                         *
 *    http://www.citeweb.net/optimum                            *
 *                                                              *
 *    (G) Gnu GPL                                               *
 ***************************************************************/




#ifndef O_SCHEDULE_H
#define O_SCHEDULE_H


/*
 * somes things to schedule the demo (timers,
 * scheduling tables, function...)
 */


#include <sys/time.h>
#include <unistd.h>


/*
 * This is the data structure holding  all usefull
 * 'times'.
 * Only One global instance of this type should be enough
 * for the demo;
 * But I moved all into a structure rather than keeping them
 * into global vars (as in previous versions) in a tentative
 * of clean programming...
 */

typedef struct{  
  struct timeval   demo_start_time;   // self-describing name...
  struct timeval   demo_current_time; // idem
  struct timeval   demo_elapsed_time; // duration since demo begining, used
                                      // for sequences transitions

  // XXX : add a "pause" field ?
  struct timeval   frame_time;        // time of last frame computation ending
  struct timeval   frame_elapsed_time;// duration between the two frames

  struct timeval   sequence_start_time;   // start-date of current sequence
  struct timeval   sequence_elapsed_time; // duration since sequence begining
} ot_time;




// update functions

void      ofn_sched_init(ot_time *Ot);    // sets  demo_start_time & frame_time
void      ofn_sched_update(ot_time *Ot);  // update demo_current_time, demo_elapsed_time,
                                // & sequence_elapsed_time
void      ofn_sched_nextframe(ot_time *Ot); // update frame_time & frame_elapsed_time
void      ofn_sched_nextsequence(ot_time *Ot); // sets sequence_start_time

// utility functions

/* converts a delta-time into a float, used for parametric
   functions based on real-time.
   simply returns secscal * (time.tv_sec + time.tv_usec/10^6)
*/
float     ofn_scale_time(float secscale,struct timeval*elapsed);  

/* compute time difference */
void      ofn_diff_timeval(struct timeval*old, struct timeval*new,
			    struct timeval*dst);

#endif
