 /***************************************************************
 *                                                              *  
 *    o_xutils.h                                                *
 *    basic Xwindow routines                                    *
 *                                                              *
 *    (G) corTeX / Optimum   1999-05-15                         *
 *    http://www.citeweb.net/optimum                            *
 *                                                              *
 *    (G) Gnu GPL                                               *
 ***************************************************************/





#ifndef O_XUTILS_H
#define O_XUTILS_H

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <sys/ipc.h> 
#include <sys/shm.h> 
#include <X11/extensions/XShm.h>


#include "o_types.h"
#include "o_std.h"
#include "o_gfxm.h"

#define True 1
#define False 0
/*
 *  Variables pour l'interface X
 */

typedef struct{
  Display      *dis;
  int           screen;
  Window        winRoot;
  int           depth;        // != bpp
  int           width;
  int           height;

  int           bpp;          // != depth
  int           rmask;
  int           gmask;
  int           bmask;
  int           byteOrder;
  int           gfxMode;      // optilib type

  int           o_XShmAvailable;       
  int           o_XShmCompletionType;  /* event type of Xshm completion event */
} ot_display;

typedef struct{
  XColor c[256];      // Optilib internal type
  Colormap colormap;  // X standard colormap type
} ot_xpalette;





/*
 * The Top window (an X-window) which will contain
 * the sub windows. 
 */
typedef struct{
  ot_display           *od;
  Window               win;
  XSetWindowAttributes winAttr;
  unsigned long        winMask;
  XSizeHints           winHint;
  
  GC                   gc;    // all-purpose gc ( initialized ok for image blitting)
  Pixmap               emptyPix; // for the sole purpose of an empty pointer
  int                  palettized; // boolean; use palette ?
  ot_xpalette          *oxp;

} ot_top_window;



/*
 * The sub windows which will contain the actual 
 * images.
 * They are children of an ot_top_window .
 */
typedef struct{
  ot_top_window  *otop;
  int             x,y;
  int             width,height;
  
} ot_sub_window;



typedef struct{
  XShmSegmentInfo      SHMInfo;
} ot_shm;

typedef struct{
  ot_display       *od;
  XImage          *xim;
  void            *data;    // alias to internal of xim or shm data (the buffer)

  int              useShm;  // Boolean
  ot_shm           os;      // if used
  uint32           flags;   // flag; bit 0 : 1 if Xshm image sent and waiting
                            // for shm_completion before next putimage
} ot_Ximage;






/********************************/
/* Preliminary workaround stuff */


// Very strange... this doesnt exists in the .h 
// of all the std headers I have seen, but it exists
// indeed in the .so library ... so here is the header...

#ifndef  XShmGetEventBase

extern int XShmGetEventBase(Display *dpy);

#endif


/*****************************/
/* Display-oriented routines */

void ofn_XOpen  (ot_display *Od, 
	       const char *Display,  // if !NULL : non-default destination display
	       int testXShm          // true if test shm availability, false if -noshm
	       );


void ofn_XClose (ot_display *Od);

/****************************/
/* Window_oriented routines */
// create a new window, create and associate a colormap
// if on 8bpp display, and create an invisible mouse pointer
void ofn_XOpenWindow(ot_display *Od, 
		     ot_top_window *Ow,                 // returns into Ow
		     int Width, int Height,         // window size
		     char *Name                     // window titlebar
		     );


void ofn_XCloseWindow(ot_top_window *Ow);

/// XXX deprecated -  use o_handler : processEvents instead
int  ofn_XKeyEvent (ot_top_window *Ow);


// ColorMap updating
void ofn_XWindowUpdateCmap(ot_top_window *Ow);
void ofn_XWindow332Cmap(ot_top_window *Ow);   // set a standard 332 colormap


/*********/
/* Image */

void ofn_XCreateImage(ot_Ximage *Oxi,
		     ot_display *Od,
		     int Width, int Height);

void ofn_XDestroyImage(ot_Ximage *Oxi);

/************************/
/* send image to window */

void ofn_XPutImage(ot_Ximage *Oxi, ot_top_window *Ow, 
		 int X, int Y,            // start into Image
		 int WX,int WY,           // destination into window
		 int Width, int Height);  // (0,0) to put all image

// use subwindow for coordinates into top-window
void ofn_XPutSimpleImage(ot_Ximage *Oxi, ot_sub_window *Ow); 



// wait for reception of the XShm completion event
void ofn_XWaitShm (ot_Ximage *Oxi);  // does not harm if no shm used (but stupid indeed) 

#endif













