 /***************************************************************
 *                                                              *  
 *    o_gfxfile.c                                               *
 *    Graphic file loading and writting                         *
 *                                                              *
 *    (G) corTeX / Optimum   1999-05-13                         *
 *    http://www.citeweb.net/optimum                            *
 *                                                              *
 *    (G) Gnu GPL                                               *
 ***************************************************************/



#include "o_gfxfile.h"


/* some utility routine that convert on-place bRGB888 to bBGRA8888;
 * used temporarlily before a true 24bpp conversion routine is 
 * available.
 */ 

static void o_o_rgb888_8(uint8 *data, int pix){
  uint8 *endd = data + 4*pix;
  uint8 *ends = data + 3*pix;
  int i;
  for (i=0; i<pix; i++){
    *(endd-4*i-1)   = 0;
    *(endd-4*i-2) = *(ends-3*i-3);
    *(endd-4*i-3) = *(ends-3*i-2);
    *(endd-4*i-4) = *(ends-3*i-1);
  }

}


// same thing : compress BGRA8888 to RGB888
// (can be on place)
static void o_o_bgra8888_8(uint8 *src, uint8 *dst, int pix){
  uint8 r,g,b;
  int i;
  for (i=0; i<pix; i++){
    r = *(src + i*4 +2);
    g = *(src + i*4 +1);
    b = *(src + i*4   );

    *(dst + i*3   ) = r;
    *(dst + i*3 +1) = g;
    *(dst + i*3 +2) = b;

  }

}



#ifndef NO_JPEG

ot_simple_surface *ofn_JpegLoad(const char *fileName, int GfxMode){
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   FILE * infile;
   JSAMPARRAY jbuffer;		/* Output row buffer */
   int row_stride;		/* physical row width in output buffer */

   uint8 *loadbuffer;    // loading buffer
   int    loaddirect;    // true if loadbuffer = OS->buffer
   ot_simple_surface *OS=NULL;
   int i;
   
   if ((infile = fopen(fileName, "rb")) == NULL) {
     return NULL;
   }


  // init structures
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);

  // specify data source
  jpeg_stdio_src(&cinfo, infile);


  (void) jpeg_read_header(&cinfo, TRUE);


  (void) jpeg_start_decompress(&cinfo);


  OS = NEW(ot_simple_surface);
  
  ofn_SimpleSurfaceCreate( OS, 
		   cinfo.output_width,cinfo.output_height,
		   GfxMode);

  if (GfxMode == MODE_bBGRA_8888 || GfxMode == MODE_RGB_888
      || GfxMode == MODE_BGR_888){
    loadbuffer = OS->buffer;
    loaddirect = True;
  } else {
    loadbuffer = malloc( 4* OS->width * OS->height);
    loaddirect = False;
  }


  // physical width (ex : 320 * RGB)
  row_stride = cinfo.output_width * cinfo.output_components;

  jbuffer = (*cinfo.mem->alloc_sarray)
    ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  i = 0;
  while (cinfo.output_scanline < cinfo.output_height) {
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
    (void) jpeg_read_scanlines(&cinfo, jbuffer, 1);
    memcpy(((char *)loadbuffer)+i*cinfo.output_width*cinfo.output_components,jbuffer[0], row_stride);
    i++;
  }

  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(infile);


  if (GfxMode == MODE_bBGRA_8888){
     o_o_rgb888_8(loadbuffer, OS->width * OS->height);
  } else if (GfxMode != MODE_bRGB_888  && GfxMode != MODE_BGR_888){  // need to fix this latter
    o_o_rgb888_8(loadbuffer, OS->width * OS->height);  // multi-pass conversion waiting for true converters

    ofn_ConvertMode( loadbuffer, OS->buffer,
		   MODE_bBGRA_8888, GfxMode,
		   OS->width * OS->height,
		   CONVERT_COMPUTE);
    
    
  }

  if (!loaddirect) free(loadbuffer);

  return (OS);
  
}



// Write a jpeg file with the image in ot_simple_surface.
// (note : jpeg is always truecolor, so palette image will be
// expanded )
void ofn_JpegWrite(const char *filename, ot_simple_surface *Os){
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   FILE * outfile;
   JSAMPROW row_pointer[1];
   uint8 *rowbuffer;       
   int src_width;
   int offset;
  


   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);
   if ((outfile = fopen(filename, "wb")) == NULL) {
     return;  // No error code, as usual
   }
   jpeg_stdio_dest(&cinfo, outfile);

   cinfo.image_width      = Os->width;
   cinfo.image_height     = Os->height;
   cinfo.input_components = 3;       // XXX if grayscale, change it to 1
   cinfo.in_color_space       = JCS_RGB ; //  JCS_GRAYSCALE.

   jpeg_set_defaults(&cinfo); // add things to change compression parameter...

   jpeg_start_compress(&cinfo, TRUE);


   // now send image to jpeglib
   /* the only big work is to convert image to fit RGB888 jpeg needs */

   rowbuffer = malloc( cinfo.input_components * cinfo.image_width );
   row_pointer[0] = rowbuffer;
   src_width = Os->width * ofn_GetBppOfMode(MODE_bBGRA_8888)/8;
   offset    = 0;


   if (Os->gfxMode == MODE_bBGRA_8888){ // only need to do BGRA8888 -> RGB888

     while (cinfo.next_scanline < cinfo.image_height) {
       o_o_bgra8888_8((uint8 *)Os->buffer + offset, rowbuffer, Os->width);
       offset += src_width;
       jpeg_write_scanlines(&cinfo, row_pointer, 1);
     }    

   } else {   // an intermediate conversion step is needed
     uint8  *interm_buffer;
     uint32 *lut;     // lookup table for palette
     
     interm_buffer = malloc( Os->width * Os->height* ofn_GetBppOfMode(MODE_bBGRA_8888));
     switch (Os->gfxMode){
     case MODE_PAL_8:
     case MODE_GRAY_8:
     case MODE_RGB_332:
       // create lut for expanding
       lut = NEW_ARRAY(uint32, 256);
       ofn_ConvertMode(Os->pRGB, lut, 
		       MODE_OT_COLOR, MODE_bBGRA_8888, 256, CONVERT_COMPUTE);
       ofn_ConvertMode(Os->buffer,
			interm_buffer,
			Os->gfxMode,
			MODE_bBGRA_8888,
			Os->width*Os->height,CONVERT_DEFAULT,
			lut);
       free(lut);
       break;
     default:
       // direct inter truecolor conversion
       ofn_ConvertMode(Os->buffer,
		       interm_buffer,
		       Os->gfxMode,
		       MODE_bBGRA_8888,
		       Os->width*Os->height,CONVERT_DEFAULT);
       break;
     }
     
     // now send lines
     while (cinfo.next_scanline < cinfo.image_height) {
       o_o_bgra8888_8(interm_buffer + offset, rowbuffer, Os->width);
       offset += src_width;
       jpeg_write_scanlines(&cinfo, row_pointer, 1);
     }    
     
     free(interm_buffer);
     
   }

   free(rowbuffer);

   // close

   jpeg_finish_compress(&cinfo);
   fclose(outfile);
   jpeg_destroy_compress(&cinfo);

}




#endif  // !NO_JPEG



#ifndef NO_PNG

#define PNG_BYTES_TO_CHECK 4
int o_o_check_if_png(const char *file_name)
{
   char buf[PNG_BYTES_TO_CHECK];
   FILE *fp;
   int res;

   /* Open the prospective PNG file. */
   if ((fp = fopen(file_name, "rb")) == NULL)
      return 0;

   /* Read in the signature bytes */
   if (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK)
     {
       fclose(fp);
       return 0;
     }

   /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature. */
   
   res = png_check_sig(buf, PNG_BYTES_TO_CHECK);
   fclose(fp);
   return (res) ;
}


ot_simple_surface *ofn_PngLoad(const char *fileName, int GfxMode){
  png_structp png_ptr;
  png_infop info_ptr;
  png_uint_32 width, height;
  int bit_depth, color_type, interlace_type;
  FILE *fp;

   png_bytep *row_pointer;
   int row;
   int rowbytes;
   float screen_gamma;

   uint8 *loadbuffer;    // loading buffer
   int    loaddirect;    // true if loadbuffer = OS->buffer
   ot_simple_surface *OS=NULL;



   if (!o_o_check_if_png(fileName))
     return (NULL);        // not a PNG file


   if ((fp = fopen(fileName, "rb")) == NULL)  // It would be very strange if it cannot open
     return NULL;                             // it twice ;)


   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply 
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                NULL, NULL, NULL);

   if (png_ptr == NULL)
   {
      fclose(fp);
      return NULL;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return NULL;
   }

   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */
   if (setjmp(png_ptr->jmpbuf))
   {
      /* Free all of the memory associated with the png_ptr and info_ptr */
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      /* If we get here, we had a problem reading the file */
      return NULL;
   }


   /* Set up the input control if you are using standard C streams */
   png_init_io(png_ptr, fp);


   /* If we have already read some of the signature */
   //   png_set_sig_bytes_read(png_ptr, sig_read);

   /* The call to png_read_info() gives us all of the information from the
    * PNG file before the first IDAT (image data chunk).  REQUIRED
    */
   png_read_info(png_ptr, info_ptr);

   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, NULL, NULL);


   OS = NEW(ot_simple_surface);
   ofn_SimpleSurfaceCreate( OS, width, height, GfxMode);



   // XXX pal 8 : assume .png file IS a palette mode
   
   if (GfxMode == MODE_bBGRA_8888 || GfxMode == MODE_PAL_8){
     loadbuffer = OS->buffer;
     loaddirect = True;
   } else {
     loadbuffer = malloc( 4* OS->width * OS->height);  // XXX on compte large la
     loaddirect = False;
   }


/**** Set up the data transformations you want.  Note that these are all
 **** optional.  Only call them if you want/need them.  Many of the
 **** transformations only work on specific types of images, and many
 **** are mutually exclusive.
 ****/

   // OUAIS
   /* tell libpng to strip 16 bit/color files down to 8 bits/color */
   png_set_strip_16(png_ptr);

   /* strip alpha bytes from the input data without combining with th
    * background (not recommended) */
   png_set_strip_alpha(png_ptr);

   /* extract multiple pixels with bit depths of 1, 2, and 4 from a single
    * byte into separate bytes (useful for paletted and grayscale images).
    */
   png_set_packing(png_ptr);

   /* change the order of packed pixels to least significant bit first
    * (not useful if you are using png_set_packing). */
   //   png_set_packswap(png_ptr);


   /* expand paletted colors into true RGB triplets when not for a
   ** PAL_8 surface */
   if ((color_type == PNG_COLOR_TYPE_PALETTE) && (GfxMode != MODE_PAL_8))
      png_set_expand(png_ptr);

   /* expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_expand(png_ptr);


    if ((color_type == PNG_COLOR_TYPE_GRAY ||
        color_type == PNG_COLOR_TYPE_GRAY_ALPHA) &&
	((GfxMode != MODE_PAL_8) && (GfxMode != MODE_GRAY_8)))
          png_set_gray_to_rgb(png_ptr);


   /* expand paletted or RGB images with transparency to full alpha channels
    * so the data will be available as RGBA quartets */
   if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_expand(png_ptr);


   screen_gamma = 2.2;  /* A good guess for PC monitors */
   //      screen_gamma = 1.7 or 1.0;  /* A good guess for Mac systems */


   /* Tell libpng to handle the gamma conversion for you.  The second call
    * is a good guess for PC generated images, but it should be configurable
    * by the user at run time by the user.  It is strongly suggested that
    * your application support gamma correction.
    */
      //   if (png_get_gAMA(png_ptr, info_ptr, &image_gamma))
      //      png_set_gamma(png_ptr, screen_gamma, image_gamma);
      //   else
   png_set_gamma(png_ptr, screen_gamma, 0.45);


   /* XXXXXX pour les 3 trucs qui suivent :
    * Jpourrais faire ca en fonction du gfxmode voulu,
    * mais comme de toute facon je convertit pour le reste,
    * ca vaut pas vraiment le coup.
    * Tout ce qu'on gagnerait c'est la place du buffer intermediaire
    */

   /* flip the RGB pixels to BGR (or RGBA to BGRA) */ 
   png_set_bgr(png_ptr);

   /* swap the RGBA or GA data to ARGB or AG (or BGRA to ABGR) */
   png_set_swap_alpha(png_ptr);

   // INutile (8bpc)
   /* swap bytes of 16 bit files to least significant byte first */
   //   png_set_swap(png_ptr);

   // OUAIS
   /* Add filler (or alpha) byte (before/after each RGB triplet) */
   png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);

   /* Turn on interlace handling.  REQUIRED if you are not using
    * png_read_image().  To see how to handle interlacing passes,
    * see the png_read_row() method below.
    */
   //   number_passes = png_set_interlace_handling(png_ptr);

   /* optional call to gamma correct and add the background to the palette
    * and update info structure.  REQUIRED if you are expecting libpng to
    * update the palette for you (ie you selected such a transform above).
    */
   png_read_update_info(png_ptr, info_ptr);

   /* allocate the memory to hold the image using the fields of info_ptr. */

   /* the easiest way to read the image */

   // XXX valid onyl if png_read_update_info() has been called
   rowbytes = png_get_rowbytes(png_ptr, info_ptr);   
   //rowbytes = width * 4;


   row_pointer = NEW_ARRAY(png_bytep,height );
   for (row = 0; row < height; row++)
     row_pointer[row] = loadbuffer+ row * rowbytes;
     

   /* Now it's time to read the image.  One of these methods is REQUIRED */

      png_read_image(png_ptr, row_pointer);

   free(row_pointer);

   /* read rest of file, and get additional chunks in info_ptr - REQUIRED */
   png_read_end(png_ptr, info_ptr);


   // read the palette
   if (GfxMode == MODE_PAL_8){
     png_colorp palette=NULL;
     int num_palette;
     int i;

     if (!png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette))
       num_palette = 0;

     if (num_palette > 256) num_palette = 256; // what
     for (i=0; i<num_palette; i++){
       OS->pRGB[i].r = palette[i].red;
       OS->pRGB[i].g = palette[i].green;
       OS->pRGB[i].b = palette[i].blue;
     }

     //     free(palette); apparently we MUST NOT free it ourself - let pnglib do it

   }



   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);


   /* close the file */
   fclose(fp);

   if (!loaddirect)
     {
       ofn_ConvertMode( loadbuffer, OS->buffer,
			MODE_bBGRA_8888, GfxMode,
			OS->width * OS->height,
			CONVERT_COMPUTE);
   
       free(loadbuffer);
     }


   return (OS);
}


#endif  // !NO_PNG


