 /***************************************************************
 *                                                              *  
 *    o_handler.c                                               *
 *    handler of (X)Events                                      *
 *                                                              *
 *    (G) corTeX / Optimum   1999-05-08                         *
 *    http://www.citeweb.net/optimum                            *
 *                                                              *
 *    (G) Gnu GPL                                               *
 ***************************************************************/

#include "o_handler.h"



/************************
 * XXX code not used yet
 */

void ofn_HandlerRegistryInit(ot_handler_registry *oreg){
   oreg->list = NULL;
   oreg->lock = 0;
}

void ofn_HandlerRegistryDestroy(ot_handler_registry *oreg){
  ot_handler_item *tmp;
  ot_handler_item *next;
  
  oreg->lock = 1;  // XXX replace it with a proper semaphore code

  tmp = oreg->list;
  while (tmp != NULL){
    next = tmp->next;
    free(tmp);
    tmp = next;
  }

  oreg->list = NULL;
  

  oreg->lock = 0; // XXX
  
}

// add an item at the head of the chained list
void ofn_HandlerRegistryAddElement(ot_handler_registry *oreg,
				   int element_type,
				   ot_anything *element){
  ot_handler_item *tmp;

  oreg->lock = 1;  // XXX

  tmp = NEW(ot_handler_item);
  tmp->element_type = element_type;
  tmp->element      = element;
  tmp->next     = oreg->list;
  tmp->previous = NULL;

  if (oreg->list != NULL)
    oreg->list->previous = tmp;
  oreg->list           = tmp;

  oreg->lock = 0; // XXX
}

// very efficient : scan all the chained list to find the element pointer...
// note : if an element is referenced twice (which is an error anyway),
// it will only remove the first occurence found.
/* Rha cool, double chained-list  code ..... grrrrrr ! */
void ofn_HandlerRegistryRemoveElement(ot_handler_registry *oreg,
				      int element_type,  // not used
				      ot_anything *element){
  ot_handler_item *tmp;

  oreg->lock = 1;  // XXX

  tmp = oreg->list;

  while (tmp != NULL) {
    if (tmp->element == element) { // ok, found
      // classical chained list code
      if (tmp->previous !=NULL)
	tmp->previous->next = tmp->next;
      else
	oreg->list = tmp->next; // new head
      if (tmp->next !=NULL)
	tmp->next->previous = tmp->previous;

      free(tmp);
      tmp = NULL; // terminate loop

    } else
      tmp = tmp->next;
  }

  oreg->lock = 0; // XXX
}



/*
 * XXX End of "code not used yet"
 ********************************/




/****************************************************
 * ofn_ProcessSurfaceListEvents() and its utility
 * function.
 ****************************************************/


// check if two rectangles overlap; used to determine which
// surfaces are concerned by a ExposeEvent
static int o_o_RectOverlap(int x1, int y1, int w1, int h1,
			   int x2, int y2, int w2, int h2){
  int xrm, xrM;  // min Max values
  int yrm, yrM;

  xrm = (x1>x2 ? x1 : x2);
  yrm = (y1>y2 ? y1 : y2);
  xrM = (x1+w1<x2+w2 ? x1+w1 : x2+w2);
  yrM = (y1+h1<y2+h2 ? y1+h1 : y2+h2);

  return ( (xrm<=xrM) && (yrm<=yrM));
  
}


void ofn_ProcessSurfaceListEvents( ot_surface_list *osurfl, ot_event *oev){
  XEvent         xev;
  ot_top_window *otop;
  Bool           result;
  int            test_exposure;  // is there any surface that requires to test Exposure events ?
  int            i;  // index of the first non-null list[] element, used to get the window/display
  
  oev->type = OPL_EV_NOTHING; // by default


  i=0;
  while ( (i<osurfl->number) && !(osurfl->flags[i] & OPL_SLIST_EXIST))
    i++;
  if (i==osurfl->number){
#ifdef OPL_DEBUG
    printf("ofn_ProcessSurfaceListEvents : List contains no valid surface \n");
#endif
    return;              // No surface in this list : no event.
  }

  if (osurfl->list[i]->ops->phyMode != OPL_PHYSURF_X){
#ifdef OPL_DEBUG
    printf("ofn_ProcessSurfaceListEvents : surface not associated with a Xwindow\n");
#endif
    return;             // surface not associated with a X window
  }

  otop = osurfl->list[i]->ops->i.x.osw->otop; // get the Xwindow surrounding the surface



  /**************************************************/
  // handle events


  // Do exposure events need to be checked ?
  test_exposure = 0;  
  for (i=0; i<osurfl->number; i++)
    if (   (osurfl->flags[i] & (OPL_SLIST_EXIST | OPL_SLIST_AUTOUPDATEEXPOSE))
	== (OPL_SLIST_EXIST | OPL_SLIST_AUTOUPDATEEXPOSE)){
      test_exposure = 1;
      break;
    }
  
  // exposure events (an obscured part of the window become visible)
  if (test_exposure)
  do{
    result = XCheckWindowEvent( otop->od->dis,
				otop->win,
				ExposureMask,
				&xev);
    if (result)
      for (i=0; i<osurfl->number; i++)
	if (   (osurfl->flags[i] & (OPL_SLIST_EXIST | OPL_SLIST_AUTOUPDATEEXPOSE))
            == (OPL_SLIST_EXIST | OPL_SLIST_AUTOUPDATEEXPOSE)){
	  if (o_o_RectOverlap(xev.xexpose.x, xev.xexpose.y, xev.xexpose.width, xev.xexpose.height,
			      osurfl->list[i]->ops->i.x.osw->x,    osurfl->list[i]->ops->i.x.osw->y,
			      osurfl->list[i]->ops->i.x.osw->width,osurfl->list[i]->ops->i.x.osw->height))
	    osurfl->flags[i] |= OPL_SLIST_UPDATED;
	}

  } while (result);



  /**************************************************/
  // return the remaining events, i.e Keyboard events.

  result = XCheckWindowEvent( otop->od->dis,
			      otop->win,
			      KeyPressMask | KeyReleaseMask,
			      &xev);
  if (!result) 
    return;     // nothing


  switch (xev.type){
  case  KeyPress  :
  case  KeyRelease :
    oev->type = OPL_EV_KEY;
    oev->ev_key.pressed = (xev.type == KeyPress);
    oev->ev_key.keysym = XKeycodeToKeysym ( otop->od->dis,
					     xev.xkey.keycode,
					     0);
    break;
    
  default:
    // what ?
  }

  

}


