 /***************************************************************
 *                                                              *  
 *    o_surface.c                                               *
 *    Virtual surface (gxf buffer)                              *
 *                                                              *
 *    (G) corTeX / Optimum   1999-05-08                         *
 *    http://www.citeweb.net/optimum                            *
 *                                                              *
 *    (G) Gnu GPL                                               *
 ***************************************************************/


#include "o_surface.h"
#include <stdlib.h>
#include <string.h>



void ofn_SimpleSurfaceCreate(ot_simple_surface *Os,
			     int Width, int Height,
			     int gfxMode){
   int size;

   size = Width*Height*ofn_GetBppOfMode(gfxMode)/8;
   Os->width      = Width;
   Os->height     = Height;
   Os->gfxMode    = gfxMode;
   Os->buffer     = malloc(size);
   memset(Os->buffer, 0, size);

   switch (Os->gfxMode){
   case MODE_PAL_8:
   case MODE_GRAY_8:             
   case MODE_RGB_332:
     Os->pRGB= NEW_ARRAY(ot_color,256);
     break;
   default:
     Os->pRGB = NULL;
   }
}

void ofn_SimpleSurfaceDestroy(ot_simple_surface *Os){
  free(Os->buffer);
  free(Os->pRGB);   // was NULL if not allocated, so OK

  Os->width = 0;
  Os->height = 0;
  Os->gfxMode = 0;
  Os->buffer = NULL;
  Os->pRGB =NULL;

}


// only fills some fields;
// specifically, does Not allocate buffer memory
// (because it can be shared into an XImage in next Attach().
// To be used, you must Attach() it... even if it is AttachNone()
void ofn_SurfaceCreate( ot_surface *Os,
		      int Width, int Height,
		      int gfxMode){
   
   Os->width      = Width;
   Os->height     = Height;
   Os->gfxMode    = gfxMode;
   Os->sameBuffer = False;
   Os->buffer     = NULL;
   Os->ops        = NEW(ot_phySurface);
   Os->ops->phyMode =  OPL_PHYSURF_NONE; // not attached yet
   Os->ops->p.pAnything = NULL;
   Os->ops->p.pRGB      = NULL;

 }


// This one Does allocate memory for the buffer.
// Such surfaces are not displayable.
/* deprecated, use an ot_simple_surface instead */
void ofn_SurfaceAttachMemory( ot_surface *Os){
  int size;

  Os->ops->phyMode = OPL_PHYSURF_RAM ;
  Os->ops->gfxMode = Os->gfxMode;
  Os->ops->palMode = PAL_NONE ;   // ?

  size = ofn_GetBppOfMode(Os->gfxMode) * Os->width * Os->height /8 ;

  Os->ops->i.RamBuffer = malloc(size);
  Os->sameBuffer = True;
  Os->buffer     = Os->ops->i.RamBuffer; 
  memset(Os->buffer, 0, size);


  switch (Os->gfxMode){
  case MODE_PAL_8:
  case MODE_GRAY_8:             
  case MODE_RGB_332:
    Os->ops->p.pRGB= NEW_ARRAY(ot_color,256);
    // allocate room for palette
    break;
  }

}



// This one creates an Ximage (ot_ximage) to interface with
// the Xwindow. If the gfxMode of the surface is the same than
// the Xdisplay, surface->buffer = ximage->buffer;
// Otherwise a new buffer is allocated in the surface,
// and conversion will take place for each screen-blitting.
// The image will be associated with a subwindow, at (x,y)
// inside the top_window.

void ofn_SurfaceAttachX( ot_surface *Os,
		       ot_top_window  *Ow,
		       int x, int y){

  Os->ops->phyMode = OPL_PHYSURF_X ;
  Os->ops->gfxMode = Ow->od->gfxMode;
  Os->ops->palMode = PAL_NONE ;   // ?
  Os->ops->p.pAnything = NULL;

  Os->ops->i.x.oxi = NEW(ot_Ximage);
  Os->ops->i.x.osw = NEW(ot_sub_window);  
  Os->ops->i.x.osw->otop  = Ow;    // fill the subwindow
  Os->ops->i.x.osw->x      = x;
  Os->ops->i.x.osw->y      = y;
  Os->ops->i.x.osw->width  = Os->width;
  Os->ops->i.x.osw->height = Os->height;

  ofn_XCreateImage(Os->ops->i.x.oxi,
		 Ow->od,
		 Os->width, Os->height);

  
  if (Os->ops->gfxMode == Os->gfxMode){
    Os->sameBuffer = True;
    Os->buffer     = Os->ops->i.x.oxi->data;  
  } else {
    int size; 
  
    Os->sameBuffer = False;

    size        = Os->width * Os->height * ofn_GetBppOfMode(Os->gfxMode)/8;
    Os->buffer  = malloc(size);
    memset(Os->buffer, 0, size);


  }



  // allocate room for conversion palette
  /* (XXX) yes, for all modes : either internal palette, or display palette*/
  Os->ops->p.pRGB= NEW_ARRAY( ot_color, 256);



  switch (Os->gfxMode){
  case MODE_PAL_8:
  case MODE_GRAY_8:              // yes... it is an intermediate version
  case MODE_RGB_332:


    // lookup tables for true color display
    switch(  ofn_GetBppOfMode(Os->ops->gfxMode)){
    case 16:
      Os->ops->p.pAnything= Os->ops->p.p16 = NEW_ARRAY(uint16, 256);
      break;
    case 24:
    case 32:
      Os->ops->p.pAnything= Os->ops->p.p32 = NEW_ARRAY(uint32, 256);
      break;

    }

    // XXX still need to name a PAL_xxx for Os->ops->palMode...

    // set default palette
    ofn_SurfaceSet332Palette(Os);
    break;

  default :
    if (Os->ops->gfxMode == MODE_PAL_8)   // XXX Hum, some kludge to say that 332 palette
      ofn_SurfaceSet332Palette(Os);       // must be used on the 8bpp display for truecolor surfaces
    break;
  }



}

// Break the links, destroy the XImage, free the local
// buffer if it exists
void ofn_SurfaceDetach( ot_surface *Os){
  
  
  switch (Os->ops->phyMode) {
  case OPL_PHYSURF_RAM:
    free(Os->ops->i.RamBuffer);
    break;
  case OPL_PHYSURF_X :
    ofn_XDestroyImage(Os->ops->i.x.oxi);
    free(Os->ops->i.x.oxi); // ximage
    free(Os->ops->i.x.osw); // subwindow
    // internal palet and LUT
    free(Os->ops->p.pRGB);      // does nothing if NULL : harmless
    free(Os->ops->p.pAnything); // does nothing if NULL : harmless
    break;
  }

  free (Os->ops);

  if (!Os->sameBuffer){
    free(Os->buffer);
  }
}

// free all the remaining stuff in the surface 
// (NULLiffy all the fields)
void ofn_SurfaceDestroy( ot_surface *Os){

  Os->width      = 0;
  Os->height     = 0;
  Os->gfxMode    = 0;
  Os->sameBuffer = False;
  Os->buffer     = NULL;
  Os->ops        = NULL;


}


/***********************/
/* Additional settings */


void ofn_SurfaceSetPalette( ot_surface *Os,  ot_color Oc[]){


  /* Now every mode can have a palette, the difference is that
   * it has not the same meaning:
   * when (Os->gfxMode == pal_8), it is the internal palette of the image.
   * when (          == truecolor), it is the Palette to be used by X
   *   when converting colors (and it will usually be a RGB_332 palette).
   *
   * So it's completely different, and yes, it's a big gross kludge.
   * XXX need to fix it for next version.
   */
  memcpy(Os->ops->p.pRGB, Oc, 256*sizeof(ot_color)); 


  switch (Os->gfxMode){

  case MODE_PAL_8:
  case MODE_GRAY_8:     
  case MODE_RGB_332:
  default :             // done for EVERY mode (see comment)




    if (Os->ops->gfxMode == MODE_PAL_8){ // direct physical palette
      int i;
      /* The cmap is sent to X automatically when surface becomes visible,
       * by ofn_SurfaceListBlit(). However I keep the update at
       * creation here, for programs that don't use SurfaceLists.
       * ( XXX this can lead to flicker at initial surfaces creations
       *  when all palettes are updated "in chaos" )
       */
      if (Os->ops->phyMode == OPL_PHYSURF_X){
	for (i=0; i<256; i++){
	  Os->ops->i.x.osw->otop->oxp->c[i].red   = Oc[i].r<<8;
	  Os->ops->i.x.osw->otop->oxp->c[i].green = Oc[i].g<<8;
	  Os->ops->i.x.osw->otop->oxp->c[i].blue  = Oc[i].b<<8;
	}
	
	ofn_XWindowUpdateCmap(Os->ops->i.x.osw->otop);
      }
      return;
    }
    // else : create a conversion palette
    // from palette to true color, using... the conversion routine
    // (pAnything == p8/16/32 , the one used in this mode)

    ofn_ConvertMode(Oc, Os->ops->p.pAnything, 
		    MODE_OT_COLOR, Os->ops->gfxMode, 256, CONVERT_COMPUTE);

    break;
    
  }

  
}


void ofn_SurfaceSet332Palette( ot_surface *Os){
  ot_color oc[256];
  int i;
  for (i=0; i<256; i++){
    oc[i].r  = (i & 0xe0) ;
    oc[i].g  = (i & 0x1c) << 3;
    oc[i].b  = (i & 0x3 ) << 6;
    oc[i].a  = 0;  // ?..
  }
  ofn_SurfaceSetPalette( Os, oc);
}



/****************/
/*   Blitting   */


// display the infamous thing on screen
void ofn_SurfaceBlit( ot_surface *Os){
  switch (Os->ops->phyMode){
  case OPL_PHYSURF_X:


    /* 1) Do conversions if needed */
    if  (Os->gfxMode==Os->ops->gfxMode){  // no conversion needed

      if (Os->sameBuffer)
	{ }  // nothing to do
      else  // copy buffer - strange ?.. 
	memcpy( Os->ops->i.x.oxi->data, 
		Os->buffer,
		Os->width * Os->height * ofn_GetBppOfMode(Os->gfxMode) /8);
    }
    else{   // there must be conversion
      
      if (ofn_GetDepthOfMode(Os->gfxMode) == 8
	  && ofn_GetDepthOfMode(Os->ops->gfxMode >8))  // expand palette-image to hicolor
	ofn_ConvertMode(Os->buffer,
			Os->ops->i.x.oxi->data,
			Os->gfxMode,
			Os->ops->gfxMode,
			Os->width*Os->height,CONVERT_DEFAULT,
			Os->ops->p.pAnything);
	  
      else
	ofn_ConvertMode(Os->buffer,
			Os->ops->i.x.oxi->data,
			Os->gfxMode,
			Os->ops->gfxMode,
			Os->width*Os->height,CONVERT_DEFAULT);

      
    } /* End of conversion test and routines */


    /* 2) display image */
    ofn_XPutSimpleImage(Os->ops->i.x.oxi, Os->ops->i.x.osw);

  
    break;
  default :
    break;
  }
}




void ofn_SurfaceListBlit( ot_surface_list *Osl, uint32 mask){
  int i;
  mask |= OPL_SLIST_EXIST; // flag.exist of course

  for (i=0; i<Osl->number; i++)
    if ((Osl->flags[i] & mask) == mask){
      ofn_SurfaceBlit(Osl->list[i]);
      if (Osl->flags[i] & OPL_SLIST_AUTODOWNDATE)
	Osl->flags[i] &= ~OPL_SLIST_UPDATED;

      /*
       * Bon alors la c'est n'importe quoi, faut tout refaire
       * Garder La palette X dans la surface pour eviter de la recalculer,
       * au lieu de la garder seulement dans la top_window
       * (pi toujours le gros caca d'une palette definie pour du trucolor,
       *  enfin bon c un peu le probleme dual koi ...)
       */

      // surface becomes visible
      if (    (mask&OPL_SLIST_VISIBLE) 
	   && !(Osl->flags[i]&OPL_SLIST_WASVISIBLE))
	   // Update palette on screen when surfaces changes
	   if ((Osl->list[i]->ops->gfxMode == MODE_PAL_8)
	       /*&& (Osl->list[i]->gfxMode == MODE_PAL_8)*/){ // direct physical palette
	     if (Osl->list[i]->ops->phyMode == OPL_PHYSURF_X){
	       int i_c;
	       for (i_c=0; i_c<256; i_c++){
		 Osl->list[i]->ops->i.x.osw->otop->oxp->c[i_c].red   = Osl->list[i]->ops->p.pRGB[i_c].r<<8;
		 Osl->list[i]->ops->i.x.osw->otop->oxp->c[i_c].green = Osl->list[i]->ops->p.pRGB[i_c].g<<8;
		 Osl->list[i]->ops->i.x.osw->otop->oxp->c[i_c].blue  = Osl->list[i]->ops->p.pRGB[i_c].b<<8;
	       }
	       ofn_XWindowUpdateCmap(Osl->list[i]->ops->i.x.osw->otop);
	     }
      
	   }
    }
  // record VISIBLE for next time
  for (i=0; i<Osl->number; i++)
    if (Osl->flags[i] & OPL_SLIST_EXIST){
      if (Osl->flags[i] & OPL_SLIST_VISIBLE)
	Osl->flags[i] |= OPL_SLIST_WASVISIBLE;
      else
	Osl->flags[i] &= ~OPL_SLIST_WASVISIBLE;
    }
      
  
  
}




ot_simple_surface *ofn_SurfaceListSnapshot( ot_surface_list *osurfl, int gfxMode, uint32 mask){
  ot_simple_surface *snap;
  int i;
  int width, height;
  int start_offset;

  switch (gfxMode){
  case MODE_PAL_8:
  case MODE_GRAY_8:
  case MODE_RGB_332:
    return (NULL);
  }


  mask |= OPL_SLIST_EXIST; // flag.exist of course
  
  height = width = 0;

  for (i=0; i<osurfl->number; i++)
    if (   (osurfl->flags[i] & OPL_SLIST_EXIST)
	 &&(osurfl->list[i]->ops->phyMode == OPL_PHYSURF_X)){
         ot_top_window *otop;
	 otop = osurfl->list[i]->ops->i.x.osw->otop;
	 width  = otop->winHint.min_width;           // hum
	 height = otop->winHint.min_height;
	 break;
    } 

  if (width != 0){
    snap = NEW(ot_simple_surface);
    ofn_SimpleSurfaceCreate(snap, width, height, gfxMode);
  }
  else return (NULL);



  memset(snap->buffer, 0, width * height * ofn_GetDepthOfMode(gfxMode)/8);

  /*
  ** XXX All of this code is very similar to the one in ofn_SurfaceBlit().
  ** Look if it can be factorised.
  */

  for (i=0; i<osurfl->number; i++)
    if ((osurfl->flags[i] & mask) == mask){
      int         free_buffer = 0;
      void       *source_buffer;
      ot_surface *Os = osurfl->list[i];      
      int width_copy;
      int scanline_copy;
      int y;

      if (Os->gfxMode == gfxMode){
	source_buffer = Os->buffer;
      }
      else
      if (Os->gfxMode == MODE_PAL_8){  // need to expand the palette
	void *palette_convert;
       // palette size
       palette_convert =  malloc(256 *  ofn_GetBppOfMode(gfxMode)/8);
       // create lookup 
       ofn_ConvertMode(Os->ops->p.pRGB, palette_convert,
		       MODE_OT_COLOR, gfxMode, 256, CONVERT_COMPUTE);
       
       source_buffer = malloc(Os->width * Os->height * ofn_GetBppOfMode(gfxMode) /8);
       free_buffer = 1;  // need to free it when it's done.
       // convert surface for final mode
       ofn_ConvertMode(Os->buffer,
		       source_buffer,
		       Os->gfxMode,
		       gfxMode,
		       Os->width*Os->height,CONVERT_DEFAULT,
		       palette_convert);
       free (palette_convert);
      }
      else { // true color -> true color
	source_buffer = malloc(Os->width * Os->height * ofn_GetBppOfMode(gfxMode) /8);
	free_buffer = 1;  // need to free it when it's done.
       // convert surface for final mode
       ofn_ConvertMode(Os->buffer,
		       source_buffer,
		       Os->gfxMode,
		       gfxMode,
		       Os->width*Os->height,CONVERT_DEFAULT);
      }

      // now, copy it to the good location in the snap surface
      width_copy    = Os->width   * ofn_GetBppOfMode(gfxMode) /8;
      scanline_copy = snap->width * ofn_GetBppOfMode(gfxMode) /8;
      start_offset  = (Os->ops->i.x.osw->x + snap->width * Os->ops->i.x.osw->y)  * ofn_GetBppOfMode(gfxMode) /8;
      for (y = 0; y < Os->height; y++)
	memcpy((uint8 *)snap->buffer  + start_offset + y*scanline_copy,
	       (uint8 *)source_buffer + y*width_copy,
	       width_copy);

      if (free_buffer)
	free(source_buffer);
      
    }


      

  return (snap);
  
}



/***********************/
/* Surface List stuff  */



// create an empty list
void ofn_SurfaceListInit(ot_surface_list *osurfl){
  int i;
  osurfl->number = 4;  // some room for standard usage without realloc
  osurfl->flags  = NEW_ARRAY(uint32, osurfl->number);
  osurfl->list   = NEW_ARRAY(ot_surface *, osurfl->number);
  for (i=0; i< osurfl->number; i++){
    osurfl->flags[i] = OPL_SLIST_NOTHING;
    osurfl->list[i]  = NULL;
  }

}

// destroy the list, but not the elements
void ofn_SurfaceListDestroy(ot_surface_list *osurfl){
  free(osurfl->flags);
  free(osurfl->list);

}

// destroy the list AND the listed surfaces
void ofn_SurfaceListDestroyAll(ot_surface_list *osurfl){
  int i;
  for (i=0; i<osurfl->number; i++)
    if (osurfl->flags[i] & OPL_SLIST_EXIST)
      ofn_SurfaceListDeleteElement(osurfl, i);

  ofn_SurfaceListDestroy(osurfl);
}



// add a surface to a list, with the specified flags (except .exist)
// return :
//  #  the number of this surface in the list
//  -1 if failed
int ofn_SurfaceListAddElement(ot_surface_list *osurfl,ot_surface *osurf, uint32 flags){
  int i;
  for (i=0; i<osurfl->number; i++)
    if (!(osurfl->flags[i] & OPL_SLIST_EXIST))
      break;

  if (i==osurfl->number){  // array full; resize it
    osurfl->flags = (uint32 *)realloc(osurfl->flags,
				      (osurfl->number+1)*sizeof(uint32));
    osurfl->list  = (ot_surface **)realloc(osurfl->list,
					   (osurfl->number+1)*sizeof(ot_surface *));

    if (  (osurfl->flags == NULL)
	  ||(osurfl->list == NULL)){
      printf("ListAdd failed\n"); 
      exit(2);                      // plutot planter que gerer les erreurs !!!
    }

    osurfl->number ++;
  }

  osurfl->flags[i] = flags | OPL_SLIST_EXIST;
  osurfl->list[i]  = osurf;

  return (i);

  
}

// remove the #nth surface from the list; doesnt detroy the surface
void ofn_SurfaceListRemoveElement(ot_surface_list *osurfl, int handle){
  if (handle >= osurfl->number || handle<0) return;
  osurfl->flags[handle] = OPL_SLIST_NOTHING;
  osurfl->list[handle] = NULL;

  // TODO doesnt shrink the array -- but it's not very important
  // 


}

// remove the #nth surface from the list AND detroy the surface
void ofn_SurfaceListDeleteElement(ot_surface_list *osurfl, int handle){
  if (handle >= osurfl->number || handle<0) return;

  if (osurfl->flags[handle] & OPL_SLIST_EXIST){
    ofn_SurfaceDetach(osurfl->list[handle]);
    ofn_SurfaceDestroy(osurfl->list[handle]);
    ofn_SurfaceListRemoveElement(osurfl, handle);
  }

}














