/*
**
**  Open a window, create a truecolor surface, blit it, close
**
*/ 



#include <stdio.h>

#include "o_std.h"
#include "o_types.h"
#include "o_xutils.h"
#include "o_surface.h"
#include "o_gfxm.h"



#define WIDTH  320
#define HEIGHT 240

void main(){
  ot_display    *ODis;
  ot_top_window *OWin;
  ot_surface    *OSurf;
  uint32        *tmpbuffer;
  int x,y;

  // allocate these structures. 
  /* (I tend to do everything dynamic rather than
   * static variables, because many structure of Optilib
   * require to.)
   * And have pointer at start eliminates the need to append "&" for
   * each function calls ;)
   */
  ODis = NEW(ot_display);
  OWin = NEW(ot_top_window);
  OSurf= NEW(ot_surface);

  // Open display and create a window

  /* "" : use default display name. You can also pass
   * explicit name "display.machine:0.0"
   * True : test for MITShm extension; False: Don't use Shm
   */

  ofn_XOpen(ODis, "",True);
  ofn_XOpenWindow(ODis, OWin, WIDTH, HEIGHT, "First Example");

  // see what Optilib has detected
  printf(" XShm extension : %s\n", 
	 ODis->o_XShmAvailable ? "In use": "Not in use");
  printf(" Display mode : %s\n",
	 ofn_GetNameOfMode(ODis->gfxMode));
  // some warning for the poor owners of 8bpp display ;)
  if (ODis->gfxMode == MODE_PAL_8)
    printf(" Note : your display is 8bpp and the result will be quite ugly...\n");
  

  // Create a truecolor surface and attach it to the X window
   ofn_SurfaceCreate( OSurf, WIDTH, HEIGHT,  MODE_ARGB_8888);
   ofn_SurfaceAttachX( OSurf, OWin, 0,0); /* coordinates are (0,0) */


  /* Now you can write in the buffer of your Surface .
   * Don't forget to cast the pointer of the buffer to the
   * appropriate type when you do pointer arithmetic or
   * dereferencing.
   * OSurf->buffer is a 'void *'. In our case we defined our
   * image to be ARGB_8888, so pixels are represented by a
   * uint32. So you can access pixels with
   * *((uint32 *)OSurf->buffer + x + y*(OSurf->width))
   */

   // As an example we will fill our surface with many colors,
   // using an intermediate pointer for clarity.

   tmpbuffer = (uint32 *)OSurf->buffer;
   for (y = 0; y< HEIGHT; y++)
     for (x = 0; x< WIDTH; x++)
       *(tmpbuffer++) =     y                 // lowest byte : blue
	                 |  (((x*2)&0xff)<<8)   // 2nd byte : green
                         |  (((x*60)&0xff00)<<8);  // 3rd byte : red


   // now we can display it
   ofn_SurfaceBlit(OSurf);
   sleep(5);    // so we have time to see something...


   // Release structures

   ofn_SurfaceDetach(OSurf);    // free stuffs allocated by attach()
   ofn_SurfaceDestroy(OSurf);   // clear remaining stuff
   free(OSurf);                 // *This is not done by Destroy*


   ofn_XCloseWindow(OWin);
   ofn_XClose(ODis);

   free(OWin);
   free(ODis);
     

}

